''' Tensorflow inception score code
Derived from https://github.com/openai/improved-gan
Code derived from tensorflow/tensorflow/models/image/imagenet/classify_image.py
THIS CODE REQUIRES TENSORFLOW 1.3 or EARLIER to run in PARALLEL BATCH MODE 

To use this code, run sample.py on your model with --sample_npz, and then 
pass the experiment name in the --experiment_name.
This code also saves pool3 stats to an npz file for FID calculation.
'''
import os
import sys
import ipdb
import time
import math
import tarfile
import argparse
import pickle as pkl
from tqdm import trange
from six.moves import urllib

import numpy as np
import tensorflow as tf


MODEL_DIR = 'tf_models'
DATA_URL = 'http://download.tensorflow.org/models/image/imagenet/inception-2015-12-05.tgz'
softmax = None


def prepare_parser():
    usage = 'Parser for TF1.3- Inception Score scripts.'
    parser = argparse.ArgumentParser(description=usage)
    parser.add_argument('--experiment_dir', type=str, default='',
                        help='Which experiment''s samples.npz file to pull and evaluate')
    return parser

def run(config):
    # Inception with TF1.3 or earlier.
    # Call this function with list of images. Each of elements should be a 
    # numpy array with values ranging from 0 to 255.
    def get_inception_score(images, splits=10):
        assert(type(images) == list)
        assert(type(images[0]) == np.ndarray)
        assert(len(images[0].shape) == 3)
        assert(np.max(images[0]) > 10)
        assert(np.min(images[0]) >= 0.0)
        inps = []
        for img in images:
            img = img.astype(np.float32)
            inps.append(np.expand_dims(img, 0))

        bs = 1

        with tf.Session() as sess:
            preds, pools = [], []
            n_batches = int(math.ceil(float(len(inps)) / float(bs)))
            for i in trange(n_batches):
                inp = inps[(i * bs):min((i + 1) * bs, len(inps))]
                inp = np.concatenate(inp, 0)
                pred, pool = sess.run([softmax, pool3], {'ExpandDims:0': inp})
                preds.append(pred)
                pools.append(pool)

            preds = np.concatenate(preds, 0)
            scores = []
            for i in range(splits):
                part = preds[(i * preds.shape[0] // splits):((i + 1) * preds.shape[0] // splits), :]
                kl = part * (np.log(part) - np.log(np.expand_dims(np.mean(part, 0), 0)))
                kl = np.mean(np.sum(kl, 1))
                scores.append(np.exp(kl))
            return np.mean(scores), np.std(scores), np.squeeze(np.concatenate(pools, 0))

    # Init inception
    def _init_inception():
        global softmax, pool3
        if not os.path.exists(MODEL_DIR):
            os.makedirs(MODEL_DIR)

        filename = DATA_URL.split('/')[-1]
        filepath = os.path.join(MODEL_DIR, filename)

        if not os.path.exists(filepath):
            def _progress(count, block_size, total_size):
                sys.stdout.write('\r>> Downloading %s %.1f%%' % (
                        filename, float(count * block_size) / float(total_size) * 100.0))
                sys.stdout.flush()
            filepath, _ = urllib.request.urlretrieve(DATA_URL, filepath, _progress)
            print()
            statinfo = os.stat(filepath)
            print('Succesfully downloaded', filename, statinfo.st_size, 'bytes.')

        tarfile.open(filepath, 'r:gz').extractall(MODEL_DIR)

        with tf.gfile.FastGFile(os.path.join(MODEL_DIR, 'classify_image_graph_def.pb'), 'rb') as f:
            graph_def = tf.GraphDef()
            graph_def.ParseFromString(f.read())
            _ = tf.import_graph_def(graph_def, name='')

        # Works with an arbitrary minibatch size.
        with tf.Session() as sess:
            pool3 = sess.graph.get_tensor_by_name('pool_3:0')
            ops = pool3.graph.get_operations()
            for op_idx, op in enumerate(ops):
                for o in op.outputs:
                    shape = o.get_shape()
                    shape = [s.value for s in shape]
                    new_shape = []
                    for j, s in enumerate(shape):
                        if s == 1 and j == 0:
                            new_shape.append(None)
                        else:
                            new_shape.append(s)
                    # o._shape = tf.TensorShape(new_shape)
                    # o.set_shape(tf.TensorShape(new_shape))
                    o.set_shape(new_shape)
            w = sess.graph.get_operation_by_name('softmax/logits/MatMul').inputs[1]
            # logits = tf.matmul(tf.squeeze(pool3), w)
            logits = tf.matmul(tf.expand_dims(tf.squeeze(pool3), 0), w)
            softmax = tf.nn.softmax(logits)

    _init_inception()

    numpy_files = [f for f in os.listdir(config['experiment_dir']) if f.endswith('.npy')]
    numpy_files = sorted(numpy_files, key=lambda s: int(s[3:-4]))

    is_means = {}
    is_stds = {}
    for fname in numpy_files:
        filepath = os.path.join(config['experiment_dir'], fname)
        print('loading {}...'.format(filepath))
        ims = np.load(filepath)
        ims = [im.transpose(1,2,0) for im in ims]
        inc_mean, inc_std, pool_activations = get_inception_score(ims, splits=10)
        # print('Saving pool to numpy file for FID calculations...')
        # np.savez(os.path.join(config['experiment_dir'], 'TF_pool.npz'),
        #          **{'pool_mean': np.mean(pool_activations,axis=0), 'pool_var': np.cov(pool_activations, rowvar=False)})
        global_iteration = int(fname[3:-4])
        is_means[global_iteration] = inc_mean
        is_stds[global_iteration] = inc_std
        print('Inception score of {:3f} +/- {:3f}.'.format(inc_mean, inc_std))

    with open(os.path.join(config['experiment_dir'], 'is.pkl'), 'wb') as f:
        pkl.dump({ 'is_means': is_means, 'is_stds': is_stds }, f)

def main():
    # parse command line and run
    parser = prepare_parser()
    config = vars(parser.parse_args())
    print(config)
    run(config)

if __name__ == '__main__':
    main()
