from torch.optim import SGD
from torch.optim import Adam
import numpy as np
import inspect
import torch
import ipdb
import re

__all__ = ['ComplexAdam', 'ComplexSGD', 'ComplexTensor']


class ComplexAdam(Adam):
    def __init__(self, params, lr=1e-3, betas=(0.9 + 0.0j, 0.999 + 0.0j), eps=1e-8, weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super(Adam, self).__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure=None):
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad
                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    # state['exp_avg'] = torch.zeros_like(p)
                    # Exponential moving average of squared gradient values
                    # state['exp_avg_sq'] = torch.zeros_like(p)

                    exp_avg = torch.zeros_like(p).cpu().numpy().astype(np.complex)  # + (0. + 0.j)
                    exp_avg = np.expand_dims(exp_avg, axis=-1) if exp_avg.ndim == 1 else exp_avg
                    exp_avg = ComplexTensor(exp_avg).to_device(p.device)
                    state['exp_avg'] = exp_avg

                    exp_avg_sq = torch.zeros_like(p).cpu().numpy().astype(np.complex)  # + (0. + 0.j)
                    exp_avg_sq = np.expand_dims(exp_avg_sq, axis=-1) if exp_avg_sq.ndim == 1 else exp_avg_sq
                    exp_avg_sq = ComplexTensor(exp_avg_sq).to_device(p.device)
                    state['exp_avg_sq'] = exp_avg_sq

                # exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad = grad.add(p, alpha=group['weight_decay'])

                grad_complex = ComplexTensor(grad)
                ipdb.set_trace()

                # Decay the first and second moment running average coefficient
                state['exp_avg'] = state['exp_avg'] * beta1 + grad_complex * (1 - beta1)
                state['exp_avg_sq'] = state['exp_avg_sq'] * beta2 + (grad * grad * (1 - beta2))
                # exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                # denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                denom = (state['exp_avg_sq'].sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                step_size = group['lr'] / bias_correction1
                p.addcdiv_(exp_avg, denom, value=-step_size)

        return loss


class ComplexSGD(SGD):
    def __init__(self, params, lr, momentum=0 + .0j, dampening=0,
                 weight_decay=0, nesterov=False):
        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov)

        super(SGD, self).__init__(params, defaults)
        assert nesterov == False

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad

                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)

                if momentum != 0:
                    if nesterov: raise NotImplementedError()

                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        # slow but it only happens once...
                        complext = torch.clone(d_p).cpu().numpy().astype(np.complex)  # + (0. + 0.j)
                        complext = np.expand_dims(complext, axis=-1) if complext.ndim == 1 else complext
                        complext = ComplexTensor(complext).to_device(d_p.device)
                        buf = param_state['momentum_buffer'] = complext
                    else:
                        buf = param_state['momentum_buffer']
                        buf = (momentum * buf) + ((d_p * (1. - dampening)).reshape(buf.shape))
                        param_state['momentum_buffer'] = buf

                    lr = group['lr']
                    step = (lr * buf)
                    d_p = 0.5 * (step + step.conjugate())
                    d_p = d_p.real.reshape(p.shape)

                    p.add_(d_p, alpha=-1)

                else:
                    raise ValueError('USE torch.optim, it is faster...')

        return loss


# Takes impiration from here:   https://github.com/williamFalcon/pytorch-complex-tensor
# but fixes a few bugs and add the conjugate operation.
class ComplexTensor(torch.Tensor):

    @staticmethod
    def __new__(cls, x, *args, **kwargs):
        # requested to init with dim list
        # double the second to last dim (..., 1, 3, 2) -> (..., 1, 6, 2)

        # reformat complex numpy arrays so we can init with them
        if isinstance(x, np.ndarray) and 'complex' in str(x.dtype):
            # collapse second to last dim
            r = x.real
            i = x.imag
            x = np.concatenate([r, i], axis=-2)
        elif isinstance(x, torch.Tensor):
            x = torch.cat([x, torch.zeros_like(x)], dim=-2)

        # x is the second to last dim in this case
        if type(x) is int and len(args) == 1:
            x = x * 2

        elif len(args) >= 2:
            size_args = list(args)
            size_args[-2] *= 2
            args = tuple(size_args)

        else:
            if isinstance(x, torch.Tensor):
                s = x.size()[-2]
            elif isinstance(x, list):
                s = len(x)
            elif isinstance(x, np.ndarray):
                s = x.shape[-2]
            if not (s % 2 == 0): raise Exception(
                'second to last dim must be even. ComplexTensor is 2 real matrices under the hood')

        # init new t
        new_t = super().__new__(cls, x, *args, **kwargs)
        return new_t

    def __deepcopy__(self, memo):
        if not self.is_leaf:
            raise RuntimeError("Only Tensors created explicitly by the user "
                               "(graph leaves) support the deepcopy protocol at the moment")
        if id(self) in memo:
            return memo[id(self)]
        with torch.no_grad():
            if self.is_sparse:
                new_tensor = self.clone()

                # hack tensor to cast as complex
                new_tensor.__class__ = ComplexTensor
            else:
                new_storage = self.storage().__deepcopy__(memo)
                new_tensor = self.new()

                # hack tensor to cast as complex
                new_tensor.__class__ = ComplexTensor
                new_tensor.set_(new_storage, self.storage_offset(), self.size(), self.stride())
            memo[id(self)] = new_tensor
            new_tensor.requires_grad = self.requires_grad
            return new_tensor

    @property
    def real(self):
        size = self.size()
        n = size[-2]
        n = n * 2
        result = self[..., :n // 2, :]
        return result

    @property
    def imag(self):
        size = self.size()
        n = size[-2]
        n = n * 2
        result = self[..., n // 2:, :]
        return result

    def __graph_copy__(self, real, imag):
        # return tensor copy but maintain graph connections
        # force the result to be a ComplexTensor
        result = torch.cat([real, imag], dim=-2)
        result.__class__ = ComplexTensor
        return result

    def conjugate(self):
        real = self.real
        imag = self.imag * -1
        return self.__graph_copy__(real, imag)

    def __add__(self, other):
        """
        Handles scalar (real, complex) and tensor (real, complex) addition
        :param other:
        :return:
        """
        real = self.real
        imag = self.imag

        # given a real tensor
        if isinstance(other, torch.Tensor) and type(other) is not ComplexTensor:
            real = real + other

        # given a complex tensor
        elif type(other) is ComplexTensor:
            real = real + other.real
            imag = imag + other.imag

        # given a real scalar
        elif np.isreal(other):
            real = real + other

        # given a complex scalar
        else:
            real = real + other.real
            imag = imag + other.imag

        return self.__graph_copy__(real, imag)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        """
        Handles scalar (real, complex) and tensor (real, complex) addition
        :param other:
        :return:
        """
        real = self.real
        imag = self.imag

        # given a real tensor
        if isinstance(other, torch.Tensor) and type(other) is not ComplexTensor:
            real = real - other

        # given a complex tensor
        elif type(other) is ComplexTensor:
            real = real - other.real
            imag = imag - other.imag

        # given a real scalar
        elif np.isreal(other):
            real = real - other

        # given a complex scalar
        else:
            real = real - other.real
            imag = imag - other.imag

        return self.__graph_copy__(real, imag)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        """
        Handles scalar (real, complex) and tensor (real, complex) multiplication
        :param other:
        :return:
        """
        real = self.real #.clone()
        imag = self.imag #.clone()

        # given a real tensor
        if isinstance(other, torch.Tensor) and type(other) is not ComplexTensor:
            real = real * other
            imag = imag * other

        # given a complex tensor
        elif type(other) is ComplexTensor:
            ac = real * other.real
            bd = imag * other.imag
            ad = real * other.imag
            bc = imag * other.real
            real = ac - bd
            imag = ad + bc

        # given a real scalar
        elif np.imag(other) == 0.0:
            real = real * np.real(other)
            imag = imag * np.real(other)

        # given a complex scalar
        else:
            ac = real * other.real
            bd = imag * other.imag
            ad = real * other.imag
            bc = imag * other.real
            real = ac - bd
            imag = ad + bc

        return self.__graph_copy__(real, imag)

    def __truediv__(self, other):
        real = self.real.clone()
        imag = self.imag.clone()

        # given a real tensor
        if isinstance(other, torch.Tensor) and type(other) is not ComplexTensor:
            raise NotImplementedError

        # given a complex tensor
        elif type(other) is ComplexTensor:
            raise NotImplementedError

        # given a real scalar
        elif np.isreal(other):
            real = real / other
            imag = imag / other

        # given a complex scalar
        else:
            raise NotImplementedError

        return self.__graph_copy__(real, imag)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.__mul__(-1)

    def mm(self, other):
        """
        Handles tensor (real, complex) matrix multiply
        :param other:
        :return:
        """
        real = self.real.clone()
        imag = self.imag.clone()

        # given a real tensor
        if isinstance(other, torch.Tensor) and type(other) is not ComplexTensor:
            real = real.mm(other)
            imag = imag.mm(other)

        # given a complex tensor
        elif type(other) is ComplexTensor:
            ac = real.mm(other.real)
            bd = imag.mm(other.imag)
            ad = real.mm(other.imag)
            bc = imag.mm(other.real)
            real = ac - bd
            imag = ad + bc

        return self.__graph_copy__(real, imag)

    def t(self):
        real = self.real.t()
        imag = self.imag.t()

        return self.__graph_copy__(real, imag)

    def abs(self):
        result = torch.sqrt(self.real ** 2 + self.imag ** 2)
        return result

    def sum(self, *args):
        real_sum = self.real.sum(*args)
        imag_sum = self.imag.sum(*args)
        return ComplexScalar(real_sum, imag_sum)

    def mean(self, *args):
        real_mean = self.real.mean(*args)
        imag_mean = self.imag.mean(*args)
        return ComplexScalar(real_mean, imag_mean)

    @property
    def grad(self):
        g = self._grad
        g.__class__ = ComplexGrad

        return g

    def cuda(self):
        real = self.real.cuda()
        imag = self.imag.cuda()

        return self.__graph_copy__(real, imag)

    def to_device(self, device):
        real = self.real.to(device)
        imag = self.imag.to(device)

        return self.__graph_copy__(real, imag)

    def __repr__(self):
        real = self.real.flatten()
        imag = self.imag.flatten()

        # use numpy to print for us
        # strings = np.asarray([f'({a}{"+" if b > 0 else "-"}{abs(b)}j)' for a, b in zip(real, imag)])
        strings = np.asarray([complex(a, b) for a, b in zip(real, imag)]).astype(np.complex64)
        strings = strings.__repr__()
        strings = re.sub('array', 'tensor', strings)
        return strings

    def __str__(self):
        return self.__repr__()

    def is_complex(self):
        return True

    def size(self, *args):
        size = self.data.size(*args)
        size = list(size)
        size[-2] = size[-2] // 2
        size = torch.Size(size)
        return size

    @property
    def shape(self):
        size = self.data.shape
        size = list(size)
        size[-2] = size[-2] // 2
        size = torch.Size(size)
        return size

    def __getitem__(self, item):

        # when real or imag is the caller return regular tensor
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        caller = calframe[1][3]

        if caller == 'real' or caller == 'imag':
            return super(ComplexTensor, self).__getitem__(item)

        # this is a regular index op, select the requested pairs then form a new ComplexTensor
        r = self.real[item]
        c = self.imag[item]

        return self.__graph_copy__(r, c)


"""
Thin wrapper for complex scalar.
Main contribution is to use only real part for backward
"""


class ComplexScalar(object):

    def __init__(self, real, imag):
        self._real = real
        self._imag = imag

    @property
    def real(self):
        return self._real

    @property
    def imag(self):
        return self._imag

    def backward(self):
        self._real.backward()

    def __repr__(self):
        return str(complex(self.real.item(), self.imag.item()))

    def __str__(self):
        return self.__repr__()


"""
Does nothing except pretty print complex grad info
"""


class ComplexGrad(torch.Tensor):

    def __deepcopy__(self, memo):
        if not self.is_leaf:
            raise RuntimeError("Only Tensors created explicitly by the user "
                               "(graph leaves) support the deepcopy protocol at the moment")
        if id(self) in memo:
            return memo[id(self)]
        with torch.no_grad():
            if self.is_sparse:
                new_tensor = self.clone()

                # hack tensor to cast as complex
                new_tensor.__class__ = ComplexGrad
            else:
                new_storage = self.storage().__deepcopy__(memo)
                new_tensor = self.new()

                # hack tensor to cast as complex
                new_tensor.__class__ = ComplexGrad
                new_tensor.set_(new_storage, self.storage_offset(), self.size(), self.stride())
            memo[id(self)] = new_tensor
            new_tensor.requires_grad = self.requires_grad
            return new_tensor

    def __repr__(self):
        size = self.size()
        split_i = size[0] // 2
        real = self[:split_i]
        imag = self[split_i:]
        size_r = real.size()

        real = real.view(-1)
        imag = imag.view(-1)

        strings = np.asarray([f'({a}{"+" if b > 0 else "-"}{abs(b)}j)' for a, b in zip(real, imag)])
        strings = strings.reshape(*size_r)
        strings = f'tensor({strings.__str__()})'
        strings = re.sub('\n', ',\n       ', strings)
        return strings

    def __str__(self):
        return self.__repr__()
